"""Fixpoint iterator over a CfG

- Transfer is a base class that must be instantiated for
each static analysis.
- fixpoint_iteration computes a fixpoint over a cfg for the given
transfer instance
"""
from cfg import *
from typing import Protocol

class Transfer[S](Protocol):
    """lists the functions that must be implemented for a static analysis over Lattice S"""
    def bottom(self) -> S:
        """the empty state, associated to unvisited nodes"""
        ...
    def init_state(self) -> S:
        """the state associated to the initial node"""
        ...
    def join(self,s1:S,s2:S) -> S:
        """join of two states when merging branches"""
        ...
    def included(self,s1: S,s2: S) -> bool:
        """return True when the first state is included in the second"""
        ...
    def tr_skip(self,s: S) -> S:
        """transfer state across a Skip transition

        Almost always transfer the state unchanged
        """
        return s
    def tr_set(self,s: S,v: str,e: ArithExpr) -> S:
        """transfer state across an assignment"""
        ...
    def tr_test(self,s: S,c: BoolExpr) -> S:
        """transfer state across a test transition"""
        ...
    def tr_err(self,s: S,e: Expr) -> S:
        """transfer state across a transition to error state"""
        ...

def fixpoint_iteration[T](transfer: Transfer[T], cfg: Cfg) -> dict[Node,T]:
    """computes the fixpoint for the given transfer functions over the given CfG
 
   output maps each node to the computed state
    """
    states: dict[Node, T] = {node: transfer.bottom() for node in cfg.g.nodes}
    states[cfg.init_node] = transfer.init_state()
    worklist: list[tuple[Node, Node, Label]] = []
    
    for dest, label in cfg.init_node.succs:
        worklist.append((cfg.init_node, dest, label))
        
    while worklist:
        src, dst, label = worklist.pop(0)
        s_src = states[src]
        if isinstance(label, LSkip):
            s_new = transfer.tr_skip(s_src)
        elif isinstance(label, LSet):
            s_new = transfer.tr_set(s_src, label.var, label.expr)
        elif isinstance(label, LTest):
            s_new = transfer.tr_test(s_src, label.cond)
        elif isinstance(label, LErr):
            s_new = transfer.tr_err(s_src, label.err)
        else:
            raise ValueError(f"Étiquette inconnue: {label}")

        s_dst_old = states[dst]
        
        if not transfer.included(s_new, s_dst_old):
            states[dst] = transfer.join(s_dst_old, s_new)

            for dest2, label2 in dst.succs:
                worklist.append((dst, dest2, label2))

    return states

"""Les noeuds non atteignables sont les noeuds que le code ne peut jamais visité ici ils sont les cas ou les tests ne sont pas vérifés ce qui est impossible à visiter parce que les tests ont la condition 1"""