"""
Operational semantics for the Why language

Provides evaluation functions for the three syntactic classes,
based on a environment mapping variables to integer values.
These functions return None when an error occur somewhere (Big step semantics)
"""

from syntax import *

def eval_aexp(env: dict[str,int], exp: ArithExpr) -> int | None:
    """evaluates an arithmetic expression safely"""
    match exp:
        case AECst(value):
            return value
        case AEVar(var):
            return env.get(var, None)
        case AEUop(uop, expr):
            val = eval_aexp(env, expr)
            if val is None:
                return None
            if uop == Uop.OPP:
                return -val
            return None
        case AEBop(bop, left_expr, right_expr):
            v1 = eval_aexp(env, left_expr)
            v2 = eval_aexp(env, right_expr)
            if v1 is None or v2 is None:
                return None
            if bop == Bop.ADD:
                return v1 + v2
            elif bop == Bop.MUL:
                return v1 * v2
            elif bop == Bop.DIV:
                if v2 == 0:
                    return None
                return v1 // v2
        case _:
            return None

def eval_bexp(env: dict[str,int], exp:BoolExpr) -> bool | None:
    """evaluates a boolean expression"""
    match exp:
        case BEPlain(aexpr):
            return eval_aexp(env,aexpr) != 0
        case BEEq(left_expr,right_expr):
            return eval_aexp(env,left_expr) == eval_aexp(env,right_expr)
        case BELeq(left_expr,right_expr):
            return eval_aexp(env,left_expr) <= eval_aexp(env,right_expr)
        case BENeg(expr):
            return not eval_bexp(env,expr)
        case _: assert False

def eval_instr(env: dict[str,int], instr: Instr) -> dict[str,int] | None:
    """evaluates an instruction"""
    match instr:
        case ISkip():
            return env
        case ISet(var,expr):
            new_env = env.copy()
            new_env[var] = eval_aexp(new_env,expr)
            return new_env
        case ISeq(first,second):
            return eval_instr(eval_instr(env,first),second)
        case ICond(cond,true_branch,false_branch):
            if eval_bexp(env,cond):
                return eval_instr(env,true_branch)
            else:
                return eval_instr(env,false_branch)
        case ILoop(cond,body):
            while eval_bexp(env,cond):
                env = eval_instr(env,body)
            return env
        case _: assert False
